from __future__ import annotations

from typing import Optional

from .base_range import BaseRange


class TemperatureRange(BaseRange):
    def __init__(self, start: float, finish: float):
        super().__init__(start, finish)

    def __repr__(self) -> str:
        return f"TemperatureRange({self.start}, {self.finish})"

    def merge(self, other: TemperatureRange) -> TemperatureRange:
        """範囲を結合します。

        Args:
            other (TemperatureRange): 結合対象。

        Returns:
            TemperatureRange: 結合した範囲。

        Example:
            >>> a = TemperatureRange(0, 10)
            >>> b = TemperatureRange(10, 20)
            >>> a.merge(b)
            TemperatureRange(0, 20)
            >>> a = TemperatureRange(10, 20)
            >>> b = TemperatureRange(0, 10)
            >>> a.merge(b)
            TemperatureRange(0, 20)

        Raises:
            ValueError: 結合可能ではない範囲が渡された場合。
        """
        if not self.mergeable(other):
            raise ValueError(
                "終了値と結合対象の開始値が同じか、"
                "開始値と結合対象の終了値が同じである必要があります。"
            )

        if self.start == other.finish:
            return TemperatureRange(other.start, self.finish)
        return TemperatureRange(self.start, other.finish)


BaseRange.register(TemperatureRange)


def is_continuous(temp_ranges_: list[TemperatureRange]) -> Optional[tuple[float, float]]:
    """範囲が連続であるかを検証します。

    Args:
        temp_ranges_ (list[HeatRange]): 温度領域のリスト。

    Returns:
        Optional[tuple[float, float]]: 領域が連続であるか。
    """
    temp_ranges = sorted(temp_ranges_)
    for i in range(len(temp_ranges)):
        if i != len(temp_ranges) - 1:
            if temp_ranges[i].finish != temp_ranges[i + 1].start:
                return temp_ranges[i].finish, temp_ranges[i + 1].start
    return None


def get_temperatures(temp_ranges_: list[TemperatureRange]) -> list[float]:
    """温度のリストを返します。

    Args:
        heat_ranges_ (list[TemperatureRange]): 温度領域のリスト。

    Returns:
        list[float]: 温度のリスト。

    Examples:
        >>> get_heats([TemperatureRange(0, 10), TemperatureRange(10, 20)])
        [0, 10, 20]
    """
    temp_ranges = sorted(temp_ranges_)
    if (values := is_continuous(temp_ranges)) is not None:
        raise ValueError(
            f'終了値と開始値が異なります。'
            f'finish: {values[0]}, '
            f'start: {values[1]}'
        )
    res: list[float] = []
    for i in range(len(temp_ranges)):
        if i != len(temp_ranges) - 1:
            if temp_ranges[i].finish != temp_ranges[i + 1].start:
                raise ValueError(
                    f'終了値と開始値が異なります。'
                    f'finish: {temp_ranges[i].finish}, '
                    f'start: {temp_ranges[i + 1].start}'
                )

        res.append(temp_ranges[i].start)
        if i == len(temp_ranges) - 1:
            res.append(temp_ranges[i].finish)

    return res


def get_temperature_ranges(temperatures_: list[float]) -> list[TemperatureRange]:
    """温度領域のリストを返します。

    Args:
        temperatures_ (list[float]): 温度のリスト。

    Returns:
        list[HeatRange]: 温度領域のリスト。

    Examples:
        >>> get_temperature_ranges([0, 10, 20])
        [TemperatureRange(0, 10), TemperatureRange(10, 20)]
    """
    temperatures = sorted(temperatures_)
    return [
        TemperatureRange(temperatures[i], temperatures[i + 1])
        for i in range(len(temperatures) - 1)
    ]


def get_temperature_transition(
    temperature_ranges: list[TemperatureRange]
) -> list[float]:
    """単調増加となるような温度の推移を返します。

    Args:
        temperature_ranges (list[TemperatureRange]): 温度領域のリスト。

    Returns:
        list[float]: 温度の推移。

    Examples:
        >>> temperature_ranges = [
                TemperatureRange(0, 10),
                TemperatureRange(20, 50),
                TemperatureRange(30, 30),
                TemperatureRange(40, 70),
                TemperatureRange(70, 70)
                TemperatureRange(70, 70)
            ]
        >>> sorted(get_temperature_transition(temperature_ranges))
        [0, 10, 20, 30, 30, 40, 50, 70, 70]
    """
    temperatures_set: set[float] = set()

    for temperature_range in temperature_ranges:
        if temperature_range.delta == 0:
            continue
        temperatures_set |= set(temperature_range())

    temperatures = list(temperatures_set)

    for temperature_range in temperature_ranges:
        if temperature_range.delta != 0:
            continue
        temp = temperature_range.start
        temp_count = temperatures.count(temp)
        if temp_count == 0:
            temperatures.extend([temp, temp])
        elif temp_count == 1:
            temperatures.extend([temp])
        elif temp_count == 2:
            # 二つ含まれる場合は何もしない。
            pass
        else:
            raise ValueError(f'同じ値が3つ以上含まれます。値: {temp_count}')

    return temperatures


def accumulate_heats(
    temperature_ranges_: list[TemperatureRange],
    temperature_range_heats: dict[TemperatureRange, float]
) -> list[float]:
    """温度領域ごとの必要熱量から全体で必要な熱量を求めます。

    Args:
        temperature_ranges_ (list[TemperatureRange]): 温度領域のリスト。
        temperature_range_heats (dict[TemperatureRange, float]):
            温度領域ごとの必要熱量。

    Returns:
        list[float]: 温度領域ごとの必要熱量を集計した結果。
    """
    temperature_ranges = sorted(temperature_ranges_)
    if temperature_ranges != (keys := sorted(list(temperature_range_heats.keys()))):
        raise ValueError(
            'temperature_range_heatsが不正です。'
            f'必要なキー: {temperature_ranges} '
            f'存在するキー: {keys}'
        )

    heats = [0.0] * (len(temperature_ranges) + 1)
    for i, temp_range in enumerate(temperature_ranges):
        heats[i + 1] = heats[i] + temperature_range_heats[temp_range]

    return heats



from collections import UserList
from copy import deepcopy
from typing import Optional, cast

from .heat_range import REL_TOL_DIGIT, HeatRange, get_heat_ranges, get_heats
from .line import Line
from .plot_segment import PlotSegment, get_plot_segments, temp_diff
from .stream import Stream
from .temperature_range import TemperatureRange


class Segment:
    heat_range: HeatRange
    heat_ranges: list[HeatRange]
    hot_temperature_range: Optional[TemperatureRange]
    cold_temmperature_range: Optional[TemperatureRange]
    hot_streams: list[Stream]
    cold_streams: list[Stream]

    hot_plot_segments: list[PlotSegment]
    cold_plot_segments: list[PlotSegment]
    hot_plot_segments_separated: list[PlotSegment]
    cold_plot_segments_separated: list[PlotSegment]
    hot_plot_segments_splitted: list[PlotSegment]
    cold_plot_segments_splitted: list[PlotSegment]

    def __init__(
        self,
        heats: tuple[float, float],
        hot_temperatures: Optional[tuple[float, float]] = None,
        cold_temperatures: Optional[tuple[float, float]] = None,
        hot_streams_: list[Stream] = [],
        cold_streams_: list[Stream] = []
    ):
        self.heat_range = HeatRange(*heats)

        self.hot_streams = deepcopy(hot_streams_)
        self.hot_temperature_range = None
        self.hot_plot_segments = []
        self.hot_plot_segments_separated_streams = []
        if hot_temperatures is not None:
            self.hot_temperature_range = TemperatureRange(*hot_temperatures)
            self.hot_plot_segments.append(PlotSegment(
                *self.heat_range(),
                *self.hot_temperature_range()
            ))
            self.hot_plot_segments_separated_streams = self.init_plot_segments_separated_streams(
                self.hot_streams,
                self.hot_temperature_range
            )

        self.cold_streams = deepcopy(cold_streams_)
        self.cold_temperature_range = None
        self.cold_plot_segments = []
        self.cold_plot_segments_separated_streams = []
        if cold_temperatures is not None:
            self.cold_temperature_range = TemperatureRange(*cold_temperatures)
            self.cold_plot_segments.append(PlotSegment(
                *self.heat_range(),
                *self.cold_temperature_range()
            ))
            self.cold_plot_segments_separated_streams = self.init_plot_segments_separated_streams(
                self.cold_streams,
                self.cold_temperature_range
            )

        hot_heats = get_heats([
            plot_segment.heat_range
            for plot_segment in self.hot_plot_segments_separated_streams
        ])
        cold_heats = get_heats([
            plot_segment.heat_range
            for plot_segment in self.cold_plot_segments_separated_streams
        ])

        # 与熱流体と受熱流体を合わせた熱量変化帯を得る。
        self.heat_ranges = get_heat_ranges(sorted(list(set(hot_heats + cold_heats))))

        self.hot_plot_segments_separated = get_plot_segments(
            self.heat_ranges,
            self.hot_plot_segments_separated_streams
        )
        self.cold_plot_segments_separated = get_plot_segments(
            self.heat_ranges,
            self.cold_plot_segments_separated_streams
        )

    def __repr__(self) -> str:
        return (
            "Segment("
            f"{self.heat_range.start}, "
            f"{self.heat_range.finish}, "
            f"{self.hot_temperature_range}, "
            f"{self.cold_temperature_range})"
        )

    def __str__(self) -> str:
        return (
            f"({self.heat_range.start}, {self.heat_range.finish})"
            f"({self.hot_temperature_range})"
            f"({self.cold_temperature_range})"
            f"{self.hot_streams}"
            f"{self.cold_streams}"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Segment):
            return NotImplemented
        return self.heat_range() == other.heat_range()

    @classmethod
    def round(self, x: float) -> float:
        return round(x, REL_TOL_DIGIT)

    def init_plot_segments_separated_streams(
        self,
        streams: list[Stream],
        temperature_range: TemperatureRange
    ) -> list[PlotSegment]:
        res: list[PlotSegment] = []
        start_heat = self.heat_range.start
        for i in range(len(streams)):
            heat = streams[i].heat()
            start_heat = self.round(start_heat)
            if i == len(streams) - 1:
                finish_heat = self.round(self.heat_range.finish)
            else:
                finish_heat = self.round(start_heat + heat)
            res.append(PlotSegment(
                start_heat,
                finish_heat,
                *temperature_range(),
                str(streams[i].id_)
            ))
            start_heat += heat
        return res

    def split(self, minimum_approach_temp_diff: float) -> None:
        """最小接近温度差を満たすように熱交換器を分割する。
        """
        minimum_temp_diff = minimum_approach_temp_diff

        hot_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
            plot_segment.heat_range: plot_segment
            for plot_segment in self.hot_plot_segments_separated
        }

        cold_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
            plot_segment.heat_range: plot_segment
            for plot_segment in self.cold_plot_segments_separated
        }

        for heat_range in self.heat_ranges:
            hot_plot_segment = hot_heat_range_plot_segment.get(heat_range, None)
            cold_plot_segment = cold_heat_range_plot_segment.get(heat_range, None)

            self.hot_temperature_range = cast(
                TemperatureRange,
                self.hot_temperature_range
            )
            self.cold_temperature_range = cast(
                TemperatureRange,
                self.cold_temperature_range
            )

            if hot_plot_segment is None or cold_plot_segment is None:
                continue

            start_temp_diff, finish_temp_diff = temp_diff(
                hot_plot_segment,
                cold_plot_segment
            )

            if start_temp_diff < minimum_temp_diff or finish_temp_diff < minimum_temp_diff:
                hot_heat_range_plot_segment[heat_range] = PlotSegment(
                    *heat_range(),
                    *self.hot_temperature_range(),
                    hot_plot_segment.uuid
                )
                cold_heat_range_plot_segment[heat_range] = PlotSegment(
                    *heat_range(),
                    *self.cold_temperature_range(),
                    cold_plot_segment.uuid
                )

                for heat_range_ in self.heat_ranges:
                    if heat_range == heat_range_:
                        continue

                    hot_plot_segment_ = hot_heat_range_plot_segment.get(
                        heat_range_,
                        None
                    )
                    cold_plot_segment_ = cold_heat_range_plot_segment.get(
                        heat_range_,
                        None
                    )
                    if hot_plot_segment_ is not None and hot_plot_segment_.uuid == hot_plot_segment.uuid:
                        hot_heat_range_plot_segment[heat_range_] = PlotSegment(
                            *heat_range_(),
                            *self.hot_temperature_range(),
                            hot_plot_segment.uuid
                        )

                    if cold_plot_segment_ is not None and cold_plot_segment_.uuid == cold_plot_segment.uuid:
                        cold_heat_range_plot_segment[heat_range_] = PlotSegment(
                            *heat_range_(),
                            *self.cold_temperature_range(),
                            cold_plot_segment.uuid
                        )

        self.hot_plot_segments_splitted = sorted(list(hot_heat_range_plot_segment.values()))
        self.cold_plot_segments_splitted = sorted(list(cold_heat_range_plot_segment.values()))


class Segments(UserList[Segment]):

    def get_lines(
        self,
        plot_segments: list[PlotSegment]
    ) -> list[Line]:
        return [plot_segment.line() for plot_segment in plot_segments]

    def hot_lines(self) -> list[Line]:
        return [line for segment in self.data for line in self.get_lines(segment.hot_plot_segments)]

    def cold_lines(self) -> list[Line]:
        return [line for segment in self.data for line in self.get_lines(segment.cold_plot_segments)]

    def hot_lines_separated(self) -> list[Line]:
        return [line for segment in self.data for line in self.get_lines(segment.hot_plot_segments_separated)]

    def cold_lines_separated(self) -> list[Line]:
        return [line for segment in self.data for line in self.get_lines(segment.cold_plot_segments_separated)]

    def hot_lines_splitted(self) -> list[Line]:
        return [line for segment in self.data for line in self.get_lines(segment.hot_plot_segments_splitted)]

    def cold_lines_splitted(self) -> list[Line]:
        return [line for segment in self.data for line in self.get_lines(segment.cold_plot_segments_splitted)]

    def split(self, minimum_approach_temp_diff: float) -> None:
        for segment in self.data:
            segment.split(minimum_approach_temp_diff)



import math
from typing import Optional

from .heat_range import HeatRange
from .plot_segment import PlotSegment


class HeatExchanger:
    """熱交換器を表すクラス。

    Attributes:
        heat_range (HeatRange): 熱交換の範囲。
        hot_stream_uuid (str): 与熱流体のid。
        cold_stream_uuid (str): 受熱流体のid。
        hot_temperature_range (TemperatureRange): 与熱流体の温度領域。
        cold_temperature_range (TemperatureRange): 受熱流体の温度領域。
        lmtd_parallel_flow (Optional[float]): 並流の場合の対数平均温度差。
        lmtd_counterflow (float): 向流の場合の対数平均温度差。
        hot_plot_segment (PlotSegment): 与熱流体のプロットセグメント。
        cold_plot_segment (PlotSegment): 受熱流体のプロットセグメント。
    """

    def __init__(
        self,
        heat_range: HeatRange,
        hot_plot_segment: PlotSegment,
        cold_plot_segment: PlotSegment
    ):
        self.heat_range = heat_range
        self.hot_plot_segment = hot_plot_segment
        self.cold_plot_segment = cold_plot_segment
        self.hot_temperature_range = self.hot_plot_segment.temperature_range
        self.cold_temperature_range = self.cold_plot_segment.temperature_range
        self.hot_stream_uuid = self.hot_plot_segment.uuid
        self.cold_stream_uuid = self.cold_plot_segment.uuid

        self.lmtd_parallel_flow = self.init_lmtd_pararell_flow()
        self.lmtd_counterflow = self.init_lmtd_counterflow()

    def __repr__(self) -> str:
        return (
            "HeatExchanger("
            f"{self.heat_range}, "
            f"{self.hot_plot_segment}, "
            f"{self.cold_plot_segment})"
        )

    def __str__(self) -> str:
        return (
            "HeatExchanger("
            f"{self.heat_range}, "
            f"{self.hot_plot_segment}, "
            f"{self.cold_plot_segment})"
        )

    def __lt__(self, other) -> bool:
        return self.heat_range < other.heat_range

    def init_lmtd_pararell_flow(self) -> Optional[float]:
        hot_low_temp, hot_high_temp = self.hot_temperature_range()
        cold_low_temp, cold_high_temp = self.cold_temperature_range()

        start_temp_diff = hot_high_temp - cold_low_temp
        finish_temp_diff = hot_low_temp - cold_high_temp

        if finish_temp_diff <= 0:
            return None

        if start_temp_diff == finish_temp_diff:
            return start_temp_diff

        return (start_temp_diff - finish_temp_diff) / math.log(start_temp_diff / finish_temp_diff)

    def init_lmtd_counterflow(self) -> float:
        hot_low_temp, hot_high_temp = self.hot_temperature_range()
        cold_low_temp, cold_high_temp = self.cold_temperature_range()

        start_temp_diff = hot_high_temp - cold_high_temp
        finish_temp_diff = hot_low_temp - cold_low_temp

        if start_temp_diff == finish_temp_diff:
            return start_temp_diff

        return (start_temp_diff - finish_temp_diff) / math.log(start_temp_diff / finish_temp_diff)


def merge_heat_exchangers(heat_exchanger: HeatExchanger, other: HeatExchanger) -> HeatExchanger:
    heat_range = heat_exchanger.heat_range.merge(other.heat_range)
    hot_plot_segment = heat_exchanger.hot_plot_segment.merge(other.hot_plot_segment)
    cold_plot_segment = heat_exchanger.cold_plot_segment.merge(other.cold_plot_segment)

    return HeatExchanger(heat_range, hot_plot_segment, cold_plot_segment)



import math
from copy import copy, deepcopy
from typing import Callable, Optional

from .heat_exchanger import HeatExchanger
from .heat_range import (REL_TOL_DIGIT, HeatRange, get_detailed_heat_ranges,
                         get_heat_ranges, get_heats)
from .plot_segment import PlotSegment, get_plot_segments, is_continuous
from .segment import Segment, Segments
from .stream import Stream, get_temperature_range_heats
from .temperature_range import (TemperatureRange, accumulate_heats,
                                get_temperatures)


def _create_composite_curve(streams: list[Stream]) -> list[PlotSegment]:
    """受け取った流体から生成した複合線を返します。

    Args:
        streams (list[Stream]): 流体のリスト。

    Returns:
        list[PlotSegment]: 複合線。
    """
    t_ranges, t_range_heats = get_temperature_range_heats(streams)
    t = get_temperatures(t_ranges)
    h = accumulate_heats(t_ranges, t_range_heats)
    return [
        PlotSegment(h[i], h[i + 1], t[i], t[i + 1]) for i in range(len(h) - 1)
    ]


def _get_heat_at_pinch_point(
    plot_segments: list[PlotSegment],
    pinch_point_temp: float,
    get_temp: Callable[[PlotSegment], float]
) -> float:
    """ピンチポイントとなる温度での熱量を返します。

    Args:
        plot_segment (list[PlotSegment]): 複合線。プロットセグメントのリスト。
        pinch_point_temp (float): ピンチポイントとなる温度[℃]。
        get_temp (Callable[[PlotSegment], float]):
            温度変化がないプロットセグメントがピンチポイントを取る場合に返す熱量。

    Returns:
        float: ピンチポイントでの熱量[W]。

    Raises:
        RuntimeError: 複合線がピンチポイントとなる温度を取らない場合。
    """
    for plot_segment in plot_segments:
        if plot_segment.contain_temperature(pinch_point_temp):
            # 熱量変化がない場合は無視する。隣接するプロットセグメントが同じ温度を取るので、
            # そのプロットセグメントの情報を利用する。
            if plot_segment.heat_range.delta == 0:
                continue

            # 与熱流体の場合、温度変化がないプロットセグメントは、右側がピンチポイントとなる。
            # 受熱流体の場合、温度変化がないプロットセグメントは、左側がピンチポイントとなる。
            if plot_segment.temperature_range.delta == 0:
                return get_temp(plot_segment)
            else:
                return plot_segment.heat_at_temperature(pinch_point_temp)

    raise RuntimeError('複合線がピンチポイントとなる温度を取りません。')


def _shift_curve(
    hot_plot_segments_: list[PlotSegment],
    cold_plot_segments_: list[PlotSegment],
    minimum_approach_temp_diff: float,
    pinch_point_temp: float
) -> tuple[list[PlotSegment], list[PlotSegment]]:
    """ピンチポイントが最小接近誤差となるように受熱複合線と与熱複合線をずらします。

    ピンチポイントは、与熱流体の場合は下に凸な箇所、受熱流体の場合は上に凸な箇所です。

    Args:
        hot_plot_segments_ (list[PlotSegment]): 与熱複合線。
        cold_plot_segments_ (list[PlotSegment]): 受熱複合線。
        minimum_approach_temp_diff (float): 最小接近温度差[℃]。
        pinch_point_temp (float): ピンチポイントの温度[℃]。

    Returns:
        tuple[list[PlotSegment], list[PlotSegment]]:
            最小接近温度差を満たす与熱複合線と受熱複合線。

    Raises:
        ValueError: 複合線が熱量に関して連続でない場合。
    """
    hot_plot_segments = sorted(deepcopy(hot_plot_segments_))
    cold_plot_segments = sorted(deepcopy(cold_plot_segments_))

    if is_continuous(hot_plot_segments) is not None \
       or is_continuous(cold_plot_segments) is not None:
        raise ValueError(
            '不正な複合線です。'
            '複合線は熱量に関して連続である必要があります。'
        )

    hot_heat_at_pinch_point = _get_heat_at_pinch_point(
        hot_plot_segments,
        pinch_point_temp,
        lambda x: x.finish_heat()
    )

    cold_heat_at_pinch_point = _get_heat_at_pinch_point(
        cold_plot_segments,
        pinch_point_temp - minimum_approach_temp_diff,
        lambda x: x.start_heat()
    )

    gap = round(
        hot_heat_at_pinch_point - cold_heat_at_pinch_point, REL_TOL_DIGIT
    )

    for plot_segment in cold_plot_segments:
        plot_segment.shift_heat(gap)

    return hot_plot_segments, cold_plot_segments


def _get_segments(
    hot_plot_segments: list[PlotSegment],
    cold_plot_segments: list[PlotSegment],
    hot_streams: list[Stream],
    cold_streams: list[Stream]
) -> Segments:
    """プロットセグメントの情報から得たセグメントを返します。

    Args:
        hot_plot_segments (list[PlotSegment]): 与熱複合線。
        cold_plot_segments (list[PlotSegment]): 受熱複合線。
        hot_streams (list[Stream]): 与熱流体。
        cold_streams (list[Stream]): 受熱流体。

    Returns:
        Segments: セグメントのリスト。
    """
    heat_ranges = get_detailed_heat_ranges(
        [
            [plot_segment.heat_range for plot_segment in hot_plot_segments],
            [plot_segment.heat_range for plot_segment in cold_plot_segments]
        ]
    )

    hot_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
        s.heat_range: s for s in get_plot_segments(heat_ranges, hot_plot_segments)
    }
    cold_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
        s.heat_range: s for s in get_plot_segments(heat_ranges, cold_plot_segments)
    }

    segments: Segments = Segments()
    for heat_range in heat_ranges:
        hot_plot_segment = hot_heat_range_plot_segment.get(heat_range, None)
        hot_temperatures: Optional[tuple[float, float]] = None
        hot_streams_: list[Stream] = []
        if hot_plot_segment is not None:
            hot_temp_range = hot_plot_segment.temperature_range
            hot_temperatures = hot_temp_range()

            for stream_ in hot_streams:
                if stream_.contain_temperatures(hot_temperatures):
                    stream = copy(stream_)
                    if hot_temp_range.delta == 0:
                        if stream.is_isothermal():
                            stream.update_heat(heat_range.delta)
                        else:
                            continue
                    else:
                        if stream.is_isothermal():
                            continue
                        else:
                            stream.update_temperature(*hot_temperatures)
                    hot_streams_.append(stream)

        cold_plot_segment = cold_heat_range_plot_segment.get(heat_range, None)
        cold_temperatures: Optional[tuple[float, float]] = None
        cold_streams_: list[Stream] = []
        if cold_plot_segment is not None:
            cold_temp_range = cold_plot_segment.temperature_range
            cold_temperatures = cold_temp_range()

            for stream_ in cold_streams:
                if stream_.contain_temperatures(cold_temperatures):
                    stream = copy(stream_)

                    if cold_temp_range.delta == 0:
                        if stream.is_isothermal():
                            stream.update_heat(heat_range.delta)
                        else:
                            continue
                    else:
                        if stream.is_isothermal():
                            continue
                        else:
                            stream.update_temperature(*cold_temperatures)
                    cold_streams_.append(stream)

        segments.append(
            Segment(
                heat_range(),
                hot_temperatures,
                cold_temperatures,
                hot_streams_,
                cold_streams_
            )
        )

    return segments


def _merge_segments(
    segments_: Segments
) -> tuple[list[PlotSegment], list[PlotSegment]]:
    """結合可能なセグメントを結合します。

    Args:
        segments_ (Segments): セグメントのリスト。

    Returns:
        tuple[list[PlotSegment], list[PlotSegment]]:
            与熱複合線と受熱複合線。
    """
    segments = deepcopy(segments_)
    hot_plot_segments = sorted([
        plot_segment for segment in segments for plot_segment in segment.hot_plot_segments_splitted
    ])
    cold_plot_segments = sorted([
        plot_segment for segment in segments for plot_segment in segment.cold_plot_segments_splitted
    ])
    heat_ranges = get_heat_ranges(
        get_heats(
            sorted([
                heat_range for segment in segments for heat_range in segment.heat_ranges
            ])
        )
    )
    heat_range_hot_plot_segment = {
        plot_segment.heat_range: plot_segment for plot_segment in hot_plot_segments
    }
    heat_range_cold_plot_segment = {
        plot_segment.heat_range: plot_segment for plot_segment in cold_plot_segments
    }

    merged_heat_ranges: list[HeatRange] = []
    merged_hot_plot_segments: list[PlotSegment] = []
    merged_cold_plot_segments: list[PlotSegment] = []
    for i in range(len(heat_ranges) - 1):
        hot_plot_segment = heat_range_hot_plot_segment.get(heat_ranges[i], None)
        cold_plot_segment = heat_range_cold_plot_segment.get(heat_ranges[i], None)
        next_hot_plot_segment = heat_range_hot_plot_segment.get(heat_ranges[i + 1], None)
        next_cold_plot_segment = heat_range_cold_plot_segment.get(heat_ranges[i + 1], None)

        if hot_plot_segment is None \
           or cold_plot_segment is None \
           or next_hot_plot_segment is None \
           or next_cold_plot_segment is None:
            continue

        if hot_plot_segment.mergiable(next_hot_plot_segment) \
           and cold_plot_segment.mergiable(next_cold_plot_segment):
            merged_heat_range = hot_plot_segment.heat_range.merge(next_hot_plot_segment.heat_range)
            merged_hot_temp_range = hot_plot_segment.temperature_range.merge(next_hot_plot_segment.temperature_range)
            merged_cold_temp_range = cold_plot_segment.temperature_range.merge(next_cold_plot_segment.temperature_range)
            merged_hot_plot_segments.append(PlotSegment(
                *merged_heat_range(),
                *merged_hot_temp_range(),
                hot_plot_segment.uuid
            ))
            merged_cold_plot_segments.append(PlotSegment(
                *merged_heat_range(),
                *merged_cold_temp_range(),
                cold_plot_segment.uuid
            ))
            merged_heat_ranges.extend([heat_ranges[i], heat_ranges[i + 1]])

    hot_plot_segments = [plot_segment for plot_segment in hot_plot_segments if plot_segment.heat_range not in merged_heat_ranges]
    cold_plot_segments = [plot_segment for plot_segment in cold_plot_segments if plot_segment.heat_range not in merged_heat_ranges]

    hot_plot_segments.extend(merged_hot_plot_segments)
    cold_plot_segments.extend(merged_cold_plot_segments)

    return sorted(hot_plot_segments), sorted(cold_plot_segments)


def get_possible_minimum_temp_diff_range(
    streams: list[Stream]
) -> TemperatureRange:
    """設定可能な最小接近温度差を返します。

    Args:
        streams (list[Stream]): 流体のリスト。

    Returns:
        float: 可能な最小接近温度差[℃]。
    """
    cold_streams = sorted(
        [stream for stream in streams if stream.is_internal() and stream.is_cold()],
        key=lambda stream: stream.input_temperature()
    )
    hot_streams = sorted(
        [stream for stream in streams if stream.is_internal() and stream.is_hot()],
        key=lambda stream: stream.output_temperature()
    )

    if not hot_streams:
        raise RuntimeError('与熱流体は少なくとも1つは指定する必要があります。')
    if not cold_streams:
        raise RuntimeError('受熱流体は少なくとも1つは指定する必要があります。')

    hot_maximum_temp = max(
        stream.input_temperature() for stream in streams if stream.is_hot()
    )

    hot_minimum_temp = min(
        stream.output_temperature() for stream in streams if stream.is_hot()
    )

    cold_maximum_temp = max(
        stream.output_temperature() for stream in streams if stream.is_cold()
    )

    cold_minimum_temp = min(
        stream.input_temperature() for stream in streams if stream.is_cold()
    )

    maximum_minimum_approch_temp_diff = min(
        hot_maximum_temp - cold_maximum_temp,
        hot_minimum_temp - cold_minimum_temp
    )

    # 与熱流体と受熱流体のセグメントを得る。
    initial_hcc = _create_composite_curve(hot_streams)
    initial_ccc = _create_composite_curve(cold_streams)

    initial_heat_ranges = get_detailed_heat_ranges(
        [
            [plot_segment.heat_range for plot_segment in initial_hcc],
            [plot_segment.heat_range for plot_segment in initial_ccc]
        ]
    )

    hcc = get_plot_segments(initial_heat_ranges, initial_hcc)
    ccc = get_plot_segments(initial_heat_ranges, initial_ccc)

    hot_maximum_heat = max(s.finish_heat() for s in hcc)
    cold_maximum_heat = max(s.finish_heat() for s in ccc)

    # 与熱複合線の方が大きい場合、複合線は右で揃える。
    # 受熱複合線の方が大きい場合、複合線は左で揃える。
    if hot_maximum_heat >= cold_maximum_heat:
        gap = hot_maximum_heat - cold_maximum_heat
        for plot_segment in ccc:
            plot_segment.shift_heat(gap)

    # ずらした複合線の与熱流体と受熱流体を合わせた熱量領域を得る。
    heat_ranges = get_detailed_heat_ranges(
        [
            [plot_segment.heat_range for plot_segment in hcc],
            [plot_segment.heat_range for plot_segment in ccc]
        ]
    )

    # 与熱流体と受熱流体を合わせた熱量領域に対応するセグメントを得る。
    hot_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
        plot_segment.heat_range: plot_segment
        for plot_segment in get_plot_segments(heat_ranges, hcc)
    }
    cold_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
        plot_segment.heat_range: plot_segment
        for plot_segment in get_plot_segments(heat_ranges, ccc)
    }

    minimum_minimum_approch_temp_diff = math.inf
    for heat_range in heat_ranges:
        hot_plot_segment = hot_heat_range_plot_segment.get(heat_range, None)
        cold_plot_segment = cold_heat_range_plot_segment.get(heat_range, None)

        if hot_plot_segment is None or cold_plot_segment is None:
            continue

        hot_start_temp, hot_finish_temp = hot_plot_segment.temperatures()
        cold_start_temp, cold_finish_temp = cold_plot_segment.temperatures()

        minimum_minimum_approch_temp_diff = min(
            minimum_minimum_approch_temp_diff,
            hot_start_temp - cold_start_temp,
            hot_finish_temp - cold_finish_temp
        )

    return TemperatureRange(
        maximum_minimum_approch_temp_diff,
        minimum_minimum_approch_temp_diff
    )


class TQDiagram:
    """TQ線図を描くために必要な情報を得るためのクラス。

    Args:
        streams (list[Stream]): 熱交換を行いたい流体。
        minimum_approach_temp_diff (float): 最小接近温度差[℃]。
        pinch_point_temp (float): ピンチポイントの温度[℃]。

    Attributes:
        heat_exchangers (list[HeatExchanger]): 熱交換器のリスト。
        hot_lines (list[Line]): TQ線図の与熱複合線。
        cold_linse (list[Line]): TQ線図の受熱複合線。
        hot_lines_separated (list[Line]): 流体ごとに分割した与熱複合線。
        cold_lines_separated (list[Line]): 流体ごとに分割した受熱複合線。
        hot_lines_splitted (list[Line]): 流体ごとに分割し、最小接近温度差を満たした与熱複合線。
        cold_lines_splitted (list[Line]): 流体ごとに分割し、最小接近温度差を満たした受熱複合線。
        hot_lines_merged (list[Line]): 熱交換器を結合した与熱複合線。
        cold_lines_merged (list[Line]): 熱交換器を結合した受熱複合線。
    """

    def __init__(
        self,
        streams: list[Stream],
        minimum_approach_temp_diff: float,
        pinch_point_temp: float
    ):
        hot_streams = sorted(
            [stream for stream in streams if stream.is_hot()],
            key=lambda stream: stream.output_temperature()
        )

        cold_streams = sorted(
            [stream for stream in streams if stream.is_cold()],
            key=lambda stream: stream.input_temperature()
        )

        # 与熱複合線と受熱複合線を得た後に、最小接近温度差を満たすようにずらす。
        hcc, ccc = _shift_curve(
            _create_composite_curve(hot_streams),
            _create_composite_curve(cold_streams),
            minimum_approach_temp_diff,
            pinch_point_temp
        )

        # セグメントを得る。
        segments = _get_segments(hcc, ccc, hot_streams, cold_streams)

        # 最小接近温度差を満たすように分割を行う
        segments.split(minimum_approach_temp_diff)

        self.hot_lines = segments.hot_lines()
        self.cold_lines = segments.cold_lines()

        self.hot_lines_separated = segments.hot_lines_separated()
        self.cold_lines_separated = segments.cold_lines_separated()

        self.hot_lines_splitted = segments.hot_lines_splitted()
        self.cold_lines_splitted = segments.cold_lines_splitted()

        # 結合可能なセグメント同士を結合する。
        hcc_merged, ccc_merged = _merge_segments(segments)
        self.hot_lines_merged = [plot_segment.line() for plot_segment in hcc_merged]
        self.cold_lines_merged = [plot_segment.line() for plot_segment in ccc_merged]

        all_heat_ranges = get_detailed_heat_ranges(
            [
                [plot_segment.heat_range for plot_segment in hcc_merged],
                [plot_segment.heat_range for plot_segment in ccc_merged]
            ]
        )
        hot_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
            s.heat_range: s for s in get_plot_segments(all_heat_ranges, hcc_merged)
        }
        cold_heat_range_plot_segment: dict[HeatRange, PlotSegment] = {
            s.heat_range: s for s in get_plot_segments(all_heat_ranges, ccc_merged)
        }

        self.heat_exchangers: list[HeatExchanger] = []
        for heat_range in all_heat_ranges:
            hot_plot_segment = hot_heat_range_plot_segment.get(heat_range, None)
            cold_plot_segment = cold_heat_range_plot_segment.get(heat_range, None)

            if hot_plot_segment is None or cold_plot_segment is None:
                continue

            self.heat_exchangers.append(
                HeatExchanger(heat_range, hot_plot_segment, cold_plot_segment)
            )



from __future__ import annotations

from itertools import chain

Line = tuple[tuple[float, float], tuple[float, float]]


def y_range(lines: list[Line]) -> tuple[float, float]:
    """xy座標系における複数の直線からyの最小値と最大値を返します。

    Args:
        lines list[Line]: 直線。

    Returns:
        tuple[float, float]:
            最小値と最大値。
    """
    return (
        min(line[0][1] for line in lines),
        max(line[1][1] for line in lines)
    )


def extract_x(lines: list[Line]) -> list[float]:
    """xy座標系における複数の直線から重複のないxの値を返します。

    Args:
        lines (list[Line]): 直線。

    Returns:
        list[float]: x座標の値。
    """
    return sorted(list(set(point[0] for point in chain(*lines))))



from __future__ import annotations

from typing import Optional

from .base_range import BaseRange

REL_TOL_DIGIT = 9


class HeatRange(BaseRange):
    def __init__(self, start: float, finish: float):
        super().__init__(start, finish)

    def __repr__(self):
        return f"HeatRange({self.start}, {self.finish})"

    def __contains__(self, temp: float, eps: float = 1e-6) -> bool:
        return self.start - eps <= temp <= self.finish + eps

    def merge(self, other: HeatRange) -> HeatRange:
        """範囲を結合します。

        Args:
            other (HeatRange): 結合対象。

        Returns:
            HeatRange: 結合した範囲。

        Example:
            >>> a = HeatRange(0, 10)
            >>> b = HeatRange(10, 20)
            >>> a.merge(b)
            HeatRange(0, 20)
            >>> a = HeatRange(10, 20)
            >>> b = HeatRange(0, 10)
            >>> a.merge(b)
            HeatRange(0, 20)

        Raises:
            ValueError: 結合可能ではない範囲が渡された場合。
        """
        if not self.mergeable(other):
            raise ValueError(
                "終了値と結合対象の開始値が同じか、"
                "開始値と結合対象の終了値が同じである必要があります。"
            )

        if self.start == other.finish:
            return HeatRange(other.start, self.finish)
        return HeatRange(self.start, other.finish)


BaseRange.register(HeatRange)


def get_heat_ranges(heats_: list[float]) -> list[HeatRange]:
    """熱量領域のリストを返します。

    Args:
        heats_ (list[float]): 熱量のリスト。

    Returns:
        list[HeatRange]: 熱量領域のリスト。

    Examples:
        >>> get_heat_ranges([0, 10, 20])
        [HeatRange(0, 10), HeatRange(10, 20)]
    """
    heats = sorted(heats_)
    return [
        HeatRange(heats[i], heats[i + 1])
        for i in range(len(heats) - 1)
    ]


def is_continuous(heat_ranges_: list[HeatRange]) -> Optional[tuple[float, float]]:
    """範囲が連続であるかを検証します。

    Args:
        heat_ranges_ (list[HeatRange]): 熱量領域のリスト。

    Returns:
        Optional[tuple[float, float]]: 領域が連続であるか。

    Examples:
        >>> is_continuous([HeatRange(0, 10), HeatRange(10, 20)])
        None
        >>> is_continuous([HeatRange(0, 10), HeatRange(15, 20)])
        (10, 15)
        >>> is_continuous([HeatRange(0, 15), HeatRange(10, 20)])
        (15, 10)
    """
    heat_ranges = sorted(heat_ranges_)
    for i in range(len(heat_ranges)):
        if i != len(heat_ranges) - 1:
            if heat_ranges[i].finish != heat_ranges[i + 1].start:
                return heat_ranges[i].finish, heat_ranges[i + 1].start
    return None


def get_heats(heat_ranges_: list[HeatRange]) -> list[float]:
    """熱量のリストを返します。

    Args:
        heat_ranges_ (list[HeatRange]): 熱量領域のリスト。

    Returns:
        list[float]: 熱量のリスト。

    Examples:
        >>> get_heats([HeatRange(0, 10), HeatRange(10, 20)])
        [0, 10, 20]
    """
    heat_ranges = sorted(heat_ranges_)
    if (values := is_continuous(heat_ranges)) is not None:
        raise ValueError(
            f'終了値と開始値が異なります。'
            f'finish: {values[0]}, '
            f'start: {values[1]}'
        )

    res: list[float] = []
    for i in range(len(heat_ranges)):
        res.append(heat_ranges[i].start)
        if i == len(heat_ranges) - 1:
            res.append(heat_ranges[i].finish)

    return res


def get_detailed_heat_ranges(
    heat_ranges_list: list[list[HeatRange]]
) -> list[HeatRange]:
    """与熱流体と受熱流体を合わせた熱量領域を返します。

    Args:
        plot_segments_list (list[list[PlotSegment]]): 複合線のリスト。

    Returns:
        list[HeatRange]: 熱量領域のリスト。
    """
    heats: set[float] = set()
    for heat_ranges in heat_ranges_list:
        heats |= set(get_heats(heat_ranges))

    return get_heat_ranges(sorted(list(heats)))



from __future__ import annotations

from abc import ABC


def minmax(a: float, b: float) -> tuple[float, float]:
    if a > b:
        return b, a
    return a, b


class BaseRange(ABC):
    """範囲を表すベースクラス。

    Args:
        start (float): 範囲の開始値。
        finish (float): 範囲の終了値。

    Attributes:
        start (float): 範囲の開始値。
        finish (float): 範囲の終了値。
        delta (float): 範囲の大きさ。
    """

    def __init__(self, start: float, finish: float):
        self.start, self.finish = minmax(start, finish)
        self.delta = self.finish - self.start

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseRange):
            return NotImplemented
        return self.start == other.start and self.finish == other.finish

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, BaseRange):
            return NotImplemented
        return not self.__eq__(other)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, BaseRange):
            return NotImplemented
        return self.start < other.start

    def __le__(self, other: object) -> bool:
        if not isinstance(other, BaseRange):
            return NotImplemented
        return self.start <= other.start

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, BaseRange):
            return NotImplemented
        return self.start > other.start

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, BaseRange):
            return NotImplemented
        return self.start >= other.start

    def __hash__(self) -> int:
        return hash((self.start, self.finish))

    def __repr__(self) -> str:
        return f"BaseRange({self.start}, {self.finish})"

    def __str__(self) -> str:
        return f"{self.start}->{self.finish}"

    def __format__(self, format_spec: str) -> str:
        return (
            f"{self.start.__format__(format_spec)}->"
            f"{self.finish.__format__(format_spec)}"
        )

    def __call__(self) -> tuple[float, float]:
        return self.start, self.finish

    def __contains__(self, value: float) -> bool:
        """範囲内にあるかを返します(閉区間)。

        Args:
            value (float): 範囲内にあるかを検証したい値。

        Returns:
            bool: 範囲内にあるかどうか。
        """
        return self.start <= value <= self.finish

    def shift(self, delta: float) -> None:
        """範囲をずらす。

        Args:
            delta (float): ずらす値
        """
        self.start += delta
        self.finish += delta

    def mergeable(self, other: BaseRange) -> bool:
        """結合可能かを返します。

        Args:
            other (BaseRange): 結合対象。

        Returns:
            bool: 結合可能かどうか。
        """
        return self.start == other.finish or self.finish == other.start



from copy import deepcopy

from .grand_composite_curve import GrandCompositeCurve
from .line import Line
from .stream import Stream
from .tq_diagram import TQDiagram, get_possible_minimum_temp_diff_range


class PinchAnalyzer:
    """エントリーポイント。
    """

    def __init__(
        self,
        streams_: list[Stream],
        minimum_approach_temp_diff: float
    ):
        hot_streams = sorted(
            [stream for stream in streams_ if stream.is_hot()],
            key=lambda s: s.output_temperature()
        )
        cold_streams = sorted(
            [stream for stream in streams_ if stream.is_cold()],
            key=lambda s: s.input_temperature()
        )

        if not hot_streams:
            raise RuntimeError('与熱流体は少なくとも1つは指定する必要があります。')
        if not cold_streams:
            raise RuntimeError('受熱流体は少なくとも1つは指定する必要があります。')

        streams = hot_streams + cold_streams
        for i, s in enumerate(streams):
            s.set_id(i + 1)

        streams = deepcopy(streams)

        minimum_approach_temp_diff_range = get_possible_minimum_temp_diff_range(streams)

        if minimum_approach_temp_diff not in minimum_approach_temp_diff_range:
            raise ValueError(
                "最小接近温度差が不正です。"
                f"指定最小接近温度差 [℃]: {minimum_approach_temp_diff}, "
                f"設定可能最小接近温度差 [℃]: {minimum_approach_temp_diff_range.start:.3f}"
                f" - {minimum_approach_temp_diff_range.finish:.3f}"
            )

        self.gcc = GrandCompositeCurve(streams, minimum_approach_temp_diff)
        id_heats = self.gcc.solve_external_heat()
        for stream in streams:
            for id_, heat in id_heats.items():
                if stream.id_ == id_:
                    stream.update_heat(heat)

        self.streams = [stream for stream in streams if stream.heat() != 0]
        self.pinch_point_temp = self.gcc.maximum_pinch_point_temp
        self.tq = TQDiagram(
            self.streams,
            minimum_approach_temp_diff,
            self.pinch_point_temp
        )

    def create_grand_composite_curve(self) -> tuple[list[float], list[float]]:
        """グランドコンポジットカーブを描くために必要な熱量と温度を返します。
        """
        return self.gcc.heats, self.gcc.temps

    def create_tq(self) -> tuple[list[Line], list[Line]]:
        """tq線図をを描くために必要な直線と熱量変化帯を返します。
        """
        return (
            self.tq.hot_lines,
            self.tq.cold_lines
        )

    def create_tq_separated(self) -> tuple[list[Line], list[Line]]:
        """流体ごとに分割したtq線図をを描くために必要な直線と熱量変化帯を返します。
        """
        return (
            self.tq.hot_lines_separated,
            self.tq.cold_lines_separated
        )

    def create_tq_splitted(self) -> tuple[list[Line], list[Line]]:
        """流体ごとに分割し、最小接近温度差の条件を満たしたtq線図をを描くために必要な直線を返します。
        """
        return (
            self.tq.hot_lines_splitted,
            self.tq.cold_lines_splitted
        )

    def create_tq_merged(self) -> tuple[list[Line], list[Line]]:
        """流体ごとに分割し、最小接近温度差の条件を満たしたtq線図をを描くために必要な直線を返します。
        """
        return (
            self.tq.hot_lines_merged,
            self.tq.cold_lines_merged
        )



from .enums import StreamType  # noqa
from .grand_composite_curve import GrandCompositeCurve  # noqa
from .line import extract_x, y_range  # noqa
from .pinch_analyzer import PinchAnalyzer  # noqa
from .stream import Stream  # noqa
from .tq_diagram import TQDiagram, get_possible_minimum_temp_diff_range  # noqa



from collections import defaultdict
from copy import deepcopy

from .stream import Stream, get_temperature_range_streams
from .temperature_range import TemperatureRange, get_temperatures


class GrandCompositeCurve:
    """グランドコンポジットカーブを作成するために必要な情報を得るためのクラス。

    Args:
        streams (list[Stream]): 熱交換を行いたい流体。
        minimum_approach_temp_diff (float): 最小接近温度差[℃]。

    Attributes:
        extarnal_streams (list[Stream]): 外部流体。
        minimum_approach_temp_diff (float): 最小接近温度差[℃]。
        maximum_pinch_point_temp (float): 最高温であるピンチポイントの温度[℃]。
        maximum_pinch_point_index (int): 最も温度が高いピンチポイントのインデックス。
        minimum_pinch_point_temp (float): 最低音であるピンチポイントの温度[℃]。
        minimum_pinch_point_index (int): 最も温度が低いピンチポイントのインデックス。
        temps (list[float]): 温度のリスト[℃]。
        heats (list[float]): 熱量のリスト[W]。
        hot_utility_target (float): 必要加熱量[W]。
        cold_utility_target (float): 必要冷却熱量[W]。
    """

    def __init__(
        self,
        streams_: list[Stream],
        minimum_approach_temp_diff: float
    ):
        streams = deepcopy(streams_)
        self.minimum_approach_temp_diff = minimum_approach_temp_diff

        # 受熱流体は最小接近温度分ずらす。
        for stream in streams:
            if stream.is_cold():
                stream.shift_temperature(self.minimum_approach_temp_diff)

        streams.sort(key=lambda stream: stream.input_temperature())

        self.external_streams = [
            stream for stream in streams if stream.is_external()
        ]

        temp_ranges, temp_range_streams = get_temperature_range_streams([
            stream for stream in streams if stream.is_internal()
        ])
        self.temps = get_temperatures(temp_ranges)
        self.heats = self._get_heats(
            temp_ranges,
            self._lacking_heats(temp_range_streams)
        )

        pinch_point_info = self._set_pinch_point()
        self.maximum_pinch_point_temp = pinch_point_info[0]
        self.maximum_pinch_point_index = pinch_point_info[1]
        self.minimum_pinch_point_temp = pinch_point_info[2]
        self.minimum_pinch_point_index = pinch_point_info[3]

        self.hot_utility_target = self.heats[-1]
        self.cold_utility_target = self.heats[0]

    def solve_external_heat(self) -> dict[int, float]:
        """外部流体による熱交換量を求めます.

        Returns:
            dict[int, float]:
                流体のidごとの交換熱量。

        Raises:
            RuntimeError: ピンチポイントを求める前に呼び出した場合。
        """
        if not hasattr(self, 'maximum_pinch_point_temp') \
           or not hasattr(self, 'minimum_pinch_point_temp'):
            raise RuntimeError('ピンチポイントが求まっていません。')

        external_cold_streams = [
            stream for stream in self.external_streams if stream.is_cold()
        ]
        external_hot_streams = [
            stream for stream in self.external_streams if stream.is_hot()
        ]

        updated_external_streams = self._update_external_streams(
            external_cold_streams,
            external_hot_streams
        )

        return {
            stream.id_: stream.heat()
            for stream in updated_external_streams
        }

    @classmethod
    def _lacking_heats(
        cls,
        temp_range_streams: defaultdict[TemperatureRange, set[Stream]]
    ) -> defaultdict[TemperatureRange, float]:
        """温度領域ごとの不足熱量を求めます.

        Args:
            temp_range_streams defaultdict[float, set[Stream]]:
                温度領域ごとの流体のセット。

        Returns:
            defaultdict[TemperatureRange, float]: 温度領域ごとの過不足熱量。
        """
        return defaultdict(int, {
            temp_range: sum([s.heat() for s in streams if s.is_hot()]) - sum([s.heat() for s in streams if s.is_cold()])
            for temp_range, streams in temp_range_streams.items()
        })

    @classmethod
    def _get_heats(
        cls,
        temp_ranges: list[TemperatureRange],
        temp_range_lacking_heat: dict[TemperatureRange, float]
    ) -> list[float]:
        """温度変化領域ごとの熱量変化を求めます。

        Args:
            temp_ranges: list[TemperatureRange]: 温度領域のリスト。
            temp_range_lacking_heat: dict[TemperatureRange, float]:
                温度領域ごとの過不足熱量。

        Returns:
            list[float]: 熱量のリスト。
        """
        temp_ranges.sort()
        heats = [0.0] * (len(temp_ranges) + 1)
        for i, temp_range in enumerate(temp_ranges):
            heats[i + 1] = heats[i] - temp_range_lacking_heat[temp_range]
        min_heat = min(heats)

        return [heat - min_heat for heat in heats]

    def _set_pinch_point(self) -> tuple[float, int, float, int]:
        """ピンチポイントとピンチポイントのインデックスを求めます。
        """
        # heat == 0は必ず存在する。
        if 0 not in self.heats:
            raise ValueError("heatに0が存在しません。")

        pinch_point_indexes = [
            i for i, heat in enumerate(self.heats) if heat == 0
        ]
        pinch_points = [self.temps[i] for i in pinch_point_indexes]

        if any(pinch_point_indexes) in [0, len(self.heats) - 1]:
            print("ピンチポイントの値が不正である可能性があります。")

        return (
            pinch_points[-1],
            pinch_point_indexes[-1],
            pinch_points[0],
            pinch_point_indexes[0]
        )

    def _update_external_streams(
            self,
            external_cold_streams: list[Stream],
            external_hot_streams: list[Stream]
    ) -> list[Stream]:
        """外部流体によって交換される熱量を求めます。

        Args:
            external_cold_streams (list[Stream]): 外部受熱流体。
            external_hot_streams (list[Stream]): 外部与熱流体。

        Returns:
            tuple[list[Stream]]:
                交換熱量の情報を追加した外部流体。
        """
        # ピンチポイントの上下に分割し、ピンチポイントに近い順に優先的に熱交換を行わせる。
        heats_heating = self.heats[self.maximum_pinch_point_index:]
        temps_heating = self.temps[self.maximum_pinch_point_index:]
        heats_cooling = self.heats[self.minimum_pinch_point_index::-1]
        temps_cooling = self.temps[self.minimum_pinch_point_index::-1]

        return (
            self._update_external_streams_heat(
                external_hot_streams,
                heats_heating,
                temps_heating,
                self.maximum_pinch_point_temp
            )
            + self._update_external_streams_heat(
                external_cold_streams,
                heats_cooling,
                temps_cooling,
                self.minimum_pinch_point_temp
            )
        )

    def _update_external_streams_heat(
        self,
        streams: list[Stream],
        heats: list[float],
        temps: list[float],
        pinch_point: float
    ) -> list[Stream]:
        """外部流体の交換熱量を決定します。

        Args:
            streams list[Stream]: 流体。
            heats list[float]: 熱量変化。
            temps list[float]: 温度変化。

        Returns:
            list[Stream]: 情報を更新した流体のリスト。
        """
        heated = 0.0
        not_heated = heats[-1]

        streams.sort(
            key=lambda stream: abs(stream.output_temperature() - pinch_point)
        )
        for stream in streams:
            # すでに交換熱量が設定されている場合にはスキップする。
            if stream.heat() != 0:
                continue

            if (stream.is_hot() and stream.output_temperature() < pinch_point) \
               or (stream.is_cold() and stream.output_temperature() > pinch_point):
                continue

            target_temperature = stream.output_temperature()
            for i in range(len(heats)):
                if i == len(heats) - 1:
                    if (stream.is_hot() and target_temperature >= temps[i]) \
                       or (stream.is_cold() and target_temperature <= temps[i]):
                        heat = not_heated
                        stream.update_heat(heat)
                        heated += heat
                        not_heated -= heat
                    break
                start_temp = temps[i]
                finish_temp = temps[i + 1]
                temp_range = TemperatureRange(start_temp, finish_temp)
                start_heat = heats[i]
                finish_heat = heats[i + 1]
                if start_heat == finish_heat:
                    continue
                if target_temperature in temp_range:
                    if temp_range.delta != 0:
                        slope = (finish_temp - start_temp) / (finish_heat - start_heat)
                        heat = 1 / slope * (target_temperature - start_temp) + start_heat
                    else:
                        heat = start_heat
                    heat = max(0, min(heat, not_heated))
                    stream.update_heat(heat)
                    heated += heat
                    not_heated -= heat
                    break

        return streams



from __future__ import annotations

import uuid
from collections.abc import Iterable
from typing import Optional

from .heat_range import HeatRange
from .heat_range import is_continuous as is_continuous_heat_ranges
from .temperature_range import TemperatureRange


class PlotSegment:
    uuid: str
    heat_range: HeatRange
    temperature_range: TemperatureRange

    def __init__(
        self,
        start_heat: float = 0.0,
        finish_heat: float = 0.0,
        start_temperature: float = 0.0,
        finish_temperature: float = 0.0,
        uuid_: Optional[str] = None
    ):
        self.heat_range = HeatRange(start_heat, finish_heat)
        self.temperature_range = TemperatureRange(start_temperature, finish_temperature)
        if uuid_ is None:
            self.uuid = str(uuid.uuid4())
        else:
            self.uuid = uuid_

    def __str__(self) -> str:
        return (
            f"heat: ({self.heat_range.start}; "
            f"{self.heat_range.finish}) "
            f"temperature: ({self.temperature_range.start}; "
            f"{self.temperature_range.finish})"
        )

    def __repr__(self) -> str:
        return (
            f"PlotSegment("
            f"{self.heat_range.start}, "
            f"{self.heat_range.finish}, "
            f"{self.temperature_range.start}, "
            f"{self.temperature_range.finish})"
        )

    def __format__(self, format_spec: str) -> str:
        return (
            f"PlotSegment("
            f"{self.heat_range.start.__format__(format_spec)}, "
            f"{self.heat_range.finish.__format__(format_spec)}, "
            f"{self.temperature_range.start}, "
            f"{self.temperature_range.finish})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PlotSegment):
            return NotImplemented
        return self.heat_range == other.heat_range \
            and self.temperature_range == other.temperature_range

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, PlotSegment):
            return NotImplemented
        return self.heat_range < other.heat_range

    def __le__(self, other: object) -> bool:
        if not isinstance(other, PlotSegment):
            return NotImplemented
        return self.heat_range <= other.heat_range

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, PlotSegment):
            return NotImplemented
        return self.heat_range > other.heat_range

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, PlotSegment):
            return NotImplemented
        return self.heat_range >= other.heat_range

    def line(self) -> tuple[tuple[float, float], tuple[float, float]]:
        return (
            (self.heats()[0], self.temperatures()[0]),
            (self.heats()[1], self.temperatures()[1])
        )

    def heats(self) -> tuple[float, float]:
        return self.heat_range()

    def start_heat(self) -> float:
        return self.heat_range.start

    def finish_heat(self) -> float:
        return self.heat_range.finish

    def temperatures(self) -> tuple[float, float]:
        return self.temperature_range()

    def start_temperature(self) -> float:
        return self.temperature_range.start

    def finish_temperature(self) -> float:
        return self.temperature_range.finish

    def contain_heat(self, heat: float) -> bool:
        return heat in self.heat_range

    def contain_heats(self, heats: Iterable[float]) -> bool:
        return all(self.contain_heat(heat) for heat in heats)

    def contain_temperature(self, temperature: float) -> bool:
        return temperature in self.temperature_range

    def temperatures_at_heats(self, heats: tuple[float, float]) -> tuple[float, float]:
        return self.temperature_at_heat(heats[0]), self.temperature_at_heat(heats[1])

    def temperature_at_heat(self, heat: float) -> float:
        if not self.contain_heat(heat):
            raise ValueError('heatを含んでいる必要があります。')

        heat_left, heat_right = self.heat_range()
        temp_left, temp_right = self.temperature_range()

        slope = (temp_right - temp_left) / (heat_right - heat_left)
        return slope * (heat - heat_left) + temp_left

    def heat_at_temperature(self, temperature: float) -> float:
        if not self.contain_temperature(temperature):
            raise ValueError('heatを含んでいる必要があります。')

        heat_left, heat_right = self.heat_range()
        temp_left, temp_right = self.temperature_range()

        slope = (temp_right - temp_left) / (heat_right - heat_left)
        return 1 / slope * (temperature - temp_left) + heat_left

    def shift_heat(self, delta: float) -> None:
        self.heat_range.shift(delta)

    def mergiable(self, other) -> bool:
        return (self.uuid == other.uuid) \
            and (self.finish_heat() == other.start_heat()) \
            and (self.finish_temperature() == other.start_temperature())

    def merge(self, other: PlotSegment) -> PlotSegment:
        if self.finish_heat() != other.start_heat():
            raise ValueError("other's start_heat must be self's finish_heat.")
        if self.finish_temperature() != other.start_temperature():
            raise ValueError("other's start_temperature must be self's finish_temperature.")
        if self.uuid != other.uuid:
            raise ValueError("other's uuid must be self's uuid.")

        return PlotSegment(
            self.start_heat(),
            other.finish_heat(),
            self.start_temperature(),
            other.finish_temperature(),
            self.uuid
        )


def temp_diff(segment: PlotSegment, other: PlotSegment) -> tuple[float, float]:
    hot_temp_start, hot_temp_finish = segment.temperatures()
    cold_temp_start, cold_temp_finish = other.temperatures()

    return (
        hot_temp_start - cold_temp_start,
        hot_temp_finish - cold_temp_finish
    )


def get_plot_segments(
    heat_ranges: list[HeatRange],
    plot_segments: list[PlotSegment]
) -> list[PlotSegment]:
    """指定された熱量領域のプロットセグメントを返す。

    Args:
        heats (list[float]): 熱量変化。
        temps (list[float]): 温度変化。
        heat_ranges (list[HeatRange]): 熱量領域のリスト。

    Returns:
        list[PlotSegment]: プロットセグメントのリスト。
    """
    res: list[PlotSegment] = []
    for heat_range in heat_ranges:
        for plot_segment in plot_segments:
            if plot_segment.contain_heats(heat_range()):
                res.append(
                    PlotSegment(
                        *heat_range(),
                        *plot_segment.temperatures_at_heats(heat_range()),
                        plot_segment.uuid
                    )
                )
    return res


def is_continuous(
    plot_segments: list[PlotSegment]
) -> Optional[tuple[float, float]]:

    """プロットセグメントが連続であるかを検証します。

    連続であった場合はNoneを返し、連続でなかった場合は、連続ではない箇所の熱量を返します。

    Args:
        plot_segments (list[PlotSegment]): プロットセグメントのリスト。

    Returns:
        Optional[tuple[float, float]]: 領域が連続であるか。
    """
    return is_continuous_heat_ranges(
        sorted([plot_segment.heat_range for plot_segment in plot_segments])
    )



import math
from collections import defaultdict
from collections.abc import Iterable
from copy import copy

from .enums import StreamType
from .errors import InvalidStreamError
from .temperature_range import (TemperatureRange, get_temperature_ranges,
                                get_temperature_transition)


class Stream:

    def __init__(
        self,
        input_temperature: float,
        output_temperature: float,
        heat_flow: float,
        type_: StreamType,
        cost: float = 0.0
    ):
        """流体を表すクラス。

        Args:
            input_temperature (float): 入り口温度
            output_temperature (float): 出口温度
            heat_flow (float): 熱量
            type_ (StreamType): 流体種
            cost (float, optional): 流体のコスト。外部流体の場合のみ設定できる。

        Raises:
            InvalidStreamError:
                入り口温度と出口温度の大小関係と流体種の関係が不正である場合。また、外部流体
                の熱量が0以外の場合、および外部流体以外の流体の熱量が0である場合。
        """
        self.id_ = 0
        self.type_ = type_

        if self.is_internal() and heat_flow == 0:
            raise InvalidStreamError(
                "外部流体でない流体の熱量の入力値は0でない必要があります。"
                f"流体の種類: {self.type_.describe()} 熱量: {heat_flow}"
            )

        if self.is_external() and heat_flow != 0:
            raise InvalidStreamError(
                "外部流体の熱量の入力値は0である必要があります。"
                f"流体の種類: {self.type_.describe()} 熱量: {heat_flow}"
            )

        if self.is_cold() and input_temperature > output_temperature:
            raise InvalidStreamError(
                "受熱流体は入り口温度が出口温度より低い必要があります。"
                f"流体の種類: {self.type_.describe()} "
                f"入り口温度: {input_temperature} "
                f"出口温度: {output_temperature}"
            )

        if self.is_hot() and input_temperature < output_temperature:
            raise InvalidStreamError(
                "与熱流体は入り口温度が出口温度より高い必要があります。"
                f"流体の種類: {self.type_.describe()} "
                f"入り口温度: {input_temperature} "
                f"出口温度: {output_temperature}"
            )

        if self.is_internal() and cost != 0:
            raise InvalidStreamError(
                "外部流体以外にはコストを設定できません。"
                f"流体の種類: {self.type_.describe()} "
                f"コスト: {cost} "
            )

        self.cost = cost
        self.temperature_range = TemperatureRange(
            input_temperature,
            output_temperature
        )
        self.heat_flow = heat_flow

    def __repr__(self) -> str:
        return (
            f"Stream({self.input_temperature()}, "
            f"{self.output_temperature()}, "
            f"{self.heat_flow}, "
            f"{self.type_}, "
            f"{self.cost})"
        )

    def __str__(self) -> str:
        return (
            f"{self.type_.describe()}, "
            f"input [℃]: {self.input_temperature()}, "
            f"output [℃]: {self.output_temperature()}, "
            f"heat flow [W]: {self.heat_flow}"
        )

    def __format__(self, format_spec: str) -> str:
        description = self.type_.describe()
        return (
            f"{description},{'':{14 - len(description)}s}"
            f"input [℃]: {self.input_temperature().__format__(format_spec)}, "
            f"output [℃]: {self.output_temperature().__format__(format_spec)}, "
            f"heat flow [W]: {self.heat_flow.__format__(format_spec)}"
        )

    def set_id(self, id_: int) -> None:
        """流体にidを設定する。

        Args:
            id_ (int): 設定するid。

        Raises:
            ValueError: idが0以下であった場合。
        """
        if id_ <= 0:
            raise ValueError('idは0より大きい必要があります。')
        self.id_ = id_

    def is_external(self) -> bool:
        """外部流体であるかを返します。
        """
        return self.type_ in [StreamType.EXTERNAL_HOT, StreamType.EXTERNAL_COLD]

    def is_internal(self) -> bool:
        """外部流体以外であるかを返します。
        """
        return not self.is_external()

    def is_hot(self) -> bool:
        """与熱流体であるかを返します。
        """
        return self.type_ in [StreamType.HOT, StreamType.EXTERNAL_HOT]

    def is_cold(self) -> bool:
        """受熱流体であるかを返します。
        """
        return not self.is_hot()

    def is_isothermal(self) -> bool:
        """温度変化がない流体かを返します。
        """
        return math.isclose(self.temperature_range.delta, 0.0)

    def input_temperature(self) -> float:
        """入り口温度を返します。
        """
        if self.is_hot():
            return self.temperature_range.finish
        return self.temperature_range.start

    def output_temperature(self) -> float:
        """出口温度を返します。
        """
        if self.is_hot():
            return self.temperature_range.start
        return self.temperature_range.finish

    def temperature(self) -> float:
        """温度変化を返します。
        """
        return self.temperature_range.delta

    def heat(self) -> float:
        """熱量を返します。
        """
        return self.heat_flow

    def temperatures(self) -> tuple[float, float]:
        """入り口温度と出口温度を返します。
        """
        return self.temperature_range()

    def shift_temperature(self, delta: float) -> None:
        """入り口温度と出口温度をずらします。

        Args:
            delta (float): ずらす値。
        """
        self.temperature_range.shift(delta)

    def contain_temperature(self, temperature: float) -> bool:
        """与えられた温度をとるかを返します。

        Args:
            temperature (float): 検証したい温度。
        """
        return temperature in self.temperature_range

    def contain_temperatures(self, temperatures: Iterable[float]) -> bool:
        """与えられた複数の温度をとるかを返します。

        全ての温度をとる場合のみTrueを返します。

        Args:
            temperatures (list[float]): 検証したい温度のリスト。
        """
        return all(map(lambda t: t in self.temperature_range, temperatures))

    def update_temperature(
        self,
        input_temperature: float,
        output_temperature: float
    ) -> None:
        """入り口温度と出口温度を更新します。

        等温流体に対しては呼び出せません。温度の値に加えて、熱量の値を元々の温度変化と新たな温
        度変化の比に従って更新します。

        Args:
            input_temperature (float): 更新する入り口温度。
            output_temperature (float): 更新する出口温度。

        Raises:
            ValueError: 等温流体に対して温度を更新しようとした場合。
        """
        if self.is_isothermal():
            raise ValueError("等温流体は入り口温度・出口温度を変更できません。")

        old_temp_delta = self.temperature_range.delta
        self.temperature_range = TemperatureRange(input_temperature, output_temperature)

        self.heat_flow = self.heat() * self.temperature_range.delta / old_temp_delta

    def update_heat(self, heat_flow: float) -> None:
        """熱量を更新します。

        入り口温度と出口温度が異なる流体に対しては呼び出せません。

        Args:
            heat_flow (float): 更新する熱量。

        Raises:
            ValueError: 等温流体以外に対して熱量を更新しようとした場合。
        """
        if self.is_internal() and not self.is_isothermal():
            raise ValueError("非等温流体は入り口温度・出口温度を変更せずに熱量を変更できません。")

        self.heat_flow = heat_flow


def get_temperature_range_streams(
    streams: list[Stream]
) -> tuple[
    list[TemperatureRange],
    defaultdict[TemperatureRange, set[Stream]]
]:
    """温度領域と、温度領域に属する流体を返します。

    Args:
        streams (list[Stream]): 流体。

    Returns:
            list[list[TemperatureRange], defaultdict[TemperatureRange, set[Stream]]:
                温度領域、温度領域に属する流体。
    """
    temperatures = sorted(get_temperature_transition([
        stream.temperature_range for stream in streams
    ]))
    temp_ranges = sorted(get_temperature_ranges(temperatures))

    temp_streams: defaultdict[float, set[Stream]] = defaultdict(set)
    for temperature in temperatures:
        for stream in streams:
            if stream.contain_temperature(temperature):
                temp_streams[temperature].add(stream)

    temp_range_streams: defaultdict[TemperatureRange, set[Stream]] = defaultdict(set)
    for temp_range in temp_ranges:
        target_streams = temp_streams[temp_range.start] & temp_streams[temp_range.finish]

        # 温度変化が0の領域の場合、温度変化がある流体を除く
        if temp_range.delta == 0:
            target_streams -= set(stream for stream in target_streams if not stream.is_isothermal())

        for target_stream in target_streams:
            stream = copy(target_stream)
            if not stream.is_isothermal():
                stream.update_temperature(temp_range.start, temp_range.finish)
            temp_range_streams[temp_range].add(stream)

    return temp_ranges, temp_range_streams


def get_temperature_range_heats(
    streams: list[Stream]
) -> tuple[list[TemperatureRange], dict[TemperatureRange, float]]:
    temp_ranges, temp_range_streams = get_temperature_range_streams(streams)
    tem_range_heats = {
        temp_range: sum(stream.heat() for stream in temp_range_streams[temp_range])
        for temp_range in temp_ranges
    }
    return temp_ranges, tem_range_heats



from enum import Enum


class StreamType(Enum):
    """流体の種類"""

    COLD = 1
    HOT = 2
    EXTERNAL_COLD = 3
    EXTERNAL_HOT = 4

    def describe(self) -> str:
        return STREAMTYPE_STR[self.name]


STREAMTYPE_STR: dict[str, str] = {
    'COLD': 'cold',
    'HOT': 'hot',
    'EXTERNAL_COLD': 'external cold',
    'EXTERNAL_HOT': 'external hot',
}
